/*---------------------------------------------------------------------------
 
  FILENAME:
        procmon.c
 
  PURPOSE:
        Provide the wview PMON packet generator entry point.
 
  REVISION HISTORY:
        Date            Engineer        Revision        Remarks
        11/20/2007      M.S. Teel       0               Original
 
  NOTES:
        
 
  LICENSE:
        Copyright (c) 2007, Mark S. Teel (mark@teel.ws)
  
        This source code is released for free distribution under the terms 
        of the GNU General Public License.
  
----------------------------------------------------------------------------*/

//  System include files
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

//  Library include files
#include <radsystem.h>

//  Local include files
#include <procmon.h>


//  static (local) memory declarations
static WVIEW_PMON_WORK procmonWork;

enum ConfigIds
{
    CFG_ID_WVIEWD               = 0,
    CFG_ID_HTMLGEND             = 1,
    CFG_ID_WVALARMD             = 2,
    CFG_ID_WVCWOPD              = 3,
    CFG_ID_WVHTTPD              = 4,
    CFG_ID_WVIEWSQLD            = 5
};

static char *configIDs[] =
{
    "wviewd",
    "htmlgend",
    "wvalarmd",
    "wvcwopd",
    "wvhttpd",
    "wviewsqld"
};

//  methods
int pmonGetProcessPid (char* pidFilePath)
{
    FILE        *pidfile;
    int         intValue;

    pidfile = fopen (pidFilePath, "r");
    if (pidfile == NULL)
    {
        radMsgLog (PRI_CATASTROPHIC, "PMON: failed to OPEN %s!", pidFilePath);
        return ERROR;
    }

    if (fscanf (pidfile, "%d", &intValue) != 1)
    {
        radMsgLog (PRI_CATASTROPHIC, "PMON: failed to READ %s!", pidFilePath);
        fclose (pidfile);
        return ERROR;
    }

    fclose (pidfile);
    return intValue;
}

static int waitForWviewDaemon (void)
{
    WVIEW_MSG_REQUEST       msg;
    char                    srcQName[QUEUE_NAME_LENGTH+1];
    UINT                    msgType;
    UINT                    length;
    void                    *recvBfr;
    int                     retVal, done = FALSE;

    // enable message reception from the radlib router for the archive path
    radMsgRouterMessageRegister (WVIEW_MSG_TYPE_ARCHIVE_PATH);

    msg.requestType = WVIEW_RQST_TYPE_ARCHIVE_PATH;

    if (radMsgRouterMessageSend (WVIEW_MSG_TYPE_REQUEST, &msg, sizeof(msg)) == ERROR)
    {
        radMsgLog (PRI_HIGH, "waitForWviewDaemon: radMsgRouterMessageSend failed!");
        radMsgRouterMessageDeregister (WVIEW_MSG_TYPE_ARCHIVE_PATH);
        return ERROR;
    }

    // now wait for the response here
    while (!done)
    {
        radUtilsSleep (50);
        
        
        if ((retVal = radQueueRecv (radProcessQueueGetID (),
                                    srcQName,
                                    &msgType,
                                    &recvBfr,
                                    &length))
            == FALSE)
        {
            continue;
        }
        else if (retVal == ERROR)
        {
            radMsgLog (PRI_STATUS, "waitForWviewDaemon: queue is closed!");
            radMsgRouterMessageDeregister (WVIEW_MSG_TYPE_ARCHIVE_PATH);
            return ERROR;
        }
    
        // is this what we want?
        if (msgType == WVIEW_MSG_TYPE_ARCHIVE_PATH)
        {
            // yes!
            done = TRUE;
        }
    
        // release the received buffer
        radBufferRls (recvBfr);
        }
    
    // disable message reception from the radlib router for the archive path
    radMsgRouterMessageDeregister (WVIEW_MSG_TYPE_ARCHIVE_PATH);

    return OK;
}

//  system initialization
static int procmonSysInit (WVIEW_PMON_WORK *work)
{
    char            devPath[256], temp[64];
    struct stat     fileData;

    //  exit if our config file is not there
    sprintf (work->configFile, "%s/%s", WVIEW_CONFIG_DIR, PMON_CONFIG_FILENAME);
    if (stat (work->configFile, &fileData) != 0)
    {
        radMsgLogInit (PROC_NAME_PMON, FALSE, TRUE);
        radMsgLog (PRI_CATASTROPHIC, "%s does not exist - exiting...\n",
                   work->configFile);
        radMsgLogExit ();
        return ERROR_ABORT;
    }
    
    //  check for our daemon's pid file, don't run if it isn't there
    sprintf (devPath, "%s/%s", WVIEW_RUN_DIR, WVD_LOCK_FILE_NAME);
    if (stat (devPath, &fileData) != 0)
    {
        radMsgLogInit (PROC_NAME_PMON, TRUE, TRUE);
        radMsgLog (PRI_CATASTROPHIC, 
                   "wview daemon lock file %s does not exist - aborting!",
                   devPath);
        radMsgLogExit ();
        return ERROR;
    }

    sprintf (work->pidFile, "%s/%s", WVIEW_RUN_DIR, PMON_LOCK_FILE_NAME);
    sprintf (work->fifoFile, "%s/dev/%s", WVIEW_RUN_DIR, PROC_NAME_PMON);
    sprintf (work->daemonQname, "%s/dev/%s", WVIEW_RUN_DIR, PROC_NAME_DAEMON);
    sprintf (work->wviewdir, "%s", WVIEW_RUN_DIR);

    //  check for our pid file, don't run if it IS there
    if (stat (work->pidFile, &fileData) == 0)
    {
        radMsgLogInit (PROC_NAME_PMON, TRUE, TRUE);
        radMsgLog (PRI_CATASTROPHIC, 
                   "lock file %s exists, older copy may be running - aborting!",
                   work->pidFile);
        radMsgLogExit ();
        return ERROR;
    }

    return OK;
}

//  system exit
static int procmonSysExit (WVIEW_PMON_WORK *work)
{
    struct stat     fileData;

    //  delete our pid file
    if (stat (work->pidFile, &fileData) == 0)
    {
        unlink (work->pidFile);
    }

    return OK;
}

static void defaultSigHandler (int signum)
{
    int         retVal;

    switch (signum)
    {
        case SIGHUP:
            // user wants us to change the verbosity setting
            retVal = wvutilsToggleVerbosity ();
            radMsgLog (PRI_STATUS, "wvpmond: SIGHUP - toggling log verbosity %s",
                       ((retVal == 0) ? "OFF" : "ON"));

            signal (signum, defaultSigHandler);
            return;

        case SIGPIPE:
            // we have a far end socket disconnection, we'll handle it in the
            // "read/write" code
            signal (signum, defaultSigHandler);
            break;

        case SIGILL:
        case SIGBUS:
        case SIGFPE:
        case SIGSEGV:
        case SIGXFSZ:
        case SIGSYS:
            // unrecoverable signal - we must exit right now!
            radMsgLog (PRI_CATASTROPHIC, "wvpmond: recv sig %d: bailing out!", signum);
            abort ();
        
        case SIGCHLD:
            wait (NULL);
            signal (signum, defaultSigHandler);
            return;

        default:
            if (procmonWork.exiting)
            {
                signal (signum, defaultSigHandler);
                return;
            }

            procmonWork.exiting = TRUE;
            radProcessSetExitFlag ();

            // can we allow the process to exit normally?
            if (!procmonWork.inMainLoop)
            {
                // NO! - we gotta bail here!
                radMsgLog (PRI_HIGH, "wvpmond: recv sig %d: exiting now!", signum);
                radMsgRouterExit ();
                procmonSysExit (&procmonWork);
                radProcessExit ();
                radSystemExit (WVIEW_SYSTEM_ID);
                exit (0);                
            }
            
            // we can allow the process to exit normally...
            radMsgLog (PRI_HIGH, "wvpmond: recv sig %d: exiting!", signum);
        
            signal (signum, defaultSigHandler);
            break;
    }
    
    return;
}

static void msgHandler
(
    char                        *srcQueueName,
    UINT                        msgType,
    void                        *msg,
    UINT                        length,
    void                        *userData
)
{
    WVIEW_MSG_POLL_RESPONSE*    respMsg;
    int                         i;
    PMON_STIM                   stim;
    
    if (msgType == WVIEW_MSG_TYPE_POLL_RESPONSE)
    {
        // process poll response:
        respMsg = (WVIEW_MSG_POLL_RESPONSE*)msg;

        // Clear the corresponding process ticks:
        for (i = 0; i < PMON_PROCESS_MAX; i ++)
        {
            if (procmonWork.process[i].pid == respMsg->pid)
            {
                // Clear his tick counter:
                procmonWork.process[i].ticks = 0;

                // Tickle the state machine
                memset (&stim, 0, sizeof(stim));
                stim.index = i;
                stim.type  = PMON_STIM_RESPONSE;
                radStatesProcess (procmonWork.process[i].stateMachine, &stim);
                return;
            }
        }
    }

    return;
}

static void evtHandler
(
    UINT        eventsRx,
    UINT        rxData,
    void        *userData
)
{
    return;
}

static void pollTimerHandler (void *parm)
{
    int                 i, mask = 0;
    WVIEW_MSG_POLL      poll;
    PMON_STIM           stim;

    // Set active process bits:
    for (i = 0; i < PMON_PROCESS_MAX; i ++)
    {
        if (procmonWork.process[i].timeout <= 0 || 
            radStatesGetState (procmonWork.process[i].stateMachine) != PMON_STATE_IDLE)
        {
            // Skip this guy
            continue;
        }

        mask = PMON_PROCESS_SET(mask,i);
        procmonWork.process[i].ticks = procmonWork.process[i].timeout;

        // Tickle the state machine
        memset (&stim, 0, sizeof(stim));
        stim.index = i;
        stim.type  = PMON_STIM_POLL;
        radStatesProcess (procmonWork.process[i].stateMachine, &stim);
    }

    poll.mask = mask;
    if (radMsgRouterMessageSend (WVIEW_MSG_TYPE_POLL, &poll, sizeof(poll)) 
        == ERROR)
    {
        radMsgLog (PRI_HIGH, "pollTimerHandler: radMsgRouterMessageSend failed!");
    }

    radProcessTimerStart (procmonWork.pollTimer, PMON_POLL_INTERVAL);
    return;
}

static void tickTimerHandler (void *parm)
{
    int             i;
    PMON_STIM       stim;

    for (i = 0; i < PMON_PROCESS_MAX; i ++)
    {
        if (procmonWork.process[i].timeout <= 0 || 
            procmonWork.process[i].ticks <= 0)
        {
            // Skip this guy
            continue;
        }

        procmonWork.process[i].ticks --;
        if (procmonWork.process[i].ticks == 0)
        {
            // Tickle the state machine
            memset (&stim, 0, sizeof(stim));
            stim.index = i;
            stim.type  = PMON_STIM_TIMEOUT;
            radStatesProcess (procmonWork.process[i].stateMachine, &stim);
        }
    }

    radProcessTimerStart (procmonWork.tickTimer, PMON_TICK_INTERVAL);
    return;
}

//  the main entry point for the wvpmond process
int main (int argc, char *argv[])
{
    void            (*alarmHandler)(int);
    int             i, retVal, intValue;
    FILE            *pidfile;
    char            instance[32], value[_MAX_PATH];
    CF_ID           configFileId;
    struct stat     fileStatus;
    char*           subString;

    memset (&procmonWork, 0, sizeof (procmonWork));

    //  initialize some system stuff first
    retVal = procmonSysInit (&procmonWork);
    if (retVal == ERROR)
    {
        radMsgLogInit (PROC_NAME_PMON, FALSE, TRUE);
        radMsgLog (PRI_CATASTROPHIC, "wvpmond init failed!\n");
        radMsgLogExit ();
        exit (1);
    }
    else if (retVal == ERROR_ABORT)
    {
        exit (2);
    }


    //  call the global radlib system init function
    if (radSystemInit (WVIEW_SYSTEM_ID) == ERROR)
    {
        radMsgLogInit (PROC_NAME_PMON, TRUE, TRUE);
        radMsgLog (PRI_CATASTROPHIC, "radSystemInit failed!");
        radMsgLogExit ();
        exit (1);
    }


    //  call the radlib process init function
    if (radProcessInit (PROC_NAME_PMON,
                        procmonWork.fifoFile,
                        PROC_NUM_TIMERS_PMON,
                        TRUE,                      // TRUE for daemon
                        msgHandler,
                        evtHandler,
                        NULL)
        == ERROR)
    {
        printf ("\nradProcessInit failed: %s\n\n", PROC_NAME_PMON);
        radSystemExit (WVIEW_SYSTEM_ID);
        exit (1);
    }

    procmonWork.myPid = getpid ();
    pidfile = fopen (procmonWork.pidFile, "w");
    if (pidfile == NULL)
    {
        radMsgLog (PRI_CATASTROPHIC, "lock file create failed!\n");
        radProcessExit ();
        radSystemExit (WVIEW_SYSTEM_ID);
        exit (1);
    }
    fprintf (pidfile, "%d", getpid ());
    fclose (pidfile);


    alarmHandler = radProcessSignalGetHandler (SIGALRM);
    radProcessSignalCatchAll (defaultSigHandler);
    radProcessSignalCatch (SIGALRM, alarmHandler);

    procmonWork.pollTimer = radProcessTimerCreate (NULL, pollTimerHandler, NULL);
    if (procmonWork.pollTimer == NULL)
    {
        radMsgLog (PRI_HIGH, "radTimerCreate failed");
        procmonSysExit (&procmonWork);
        radProcessExit ();
        radSystemExit (WVIEW_SYSTEM_ID);
        exit (1);
    }

    procmonWork.tickTimer = radProcessTimerCreate (NULL, tickTimerHandler, NULL);
    if (procmonWork.tickTimer == NULL)
    {
        radMsgLog (PRI_HIGH, "radTimerCreate failed");
        procmonSysExit (&procmonWork);
        radProcessExit ();
        radSystemExit (WVIEW_SYSTEM_ID);
        exit (1);
    }

    //  register with the radlib message router
    if (radMsgRouterInit (WVIEW_RUN_DIR) == ERROR)
    {
        radMsgLog (PRI_HIGH, "radMsgRouterInit failed!");
        procmonSysExit (&procmonWork);
        radProcessExit ();
        radSystemExit (WVIEW_SYSTEM_ID);
        exit (1);
    }

    
    // wait for the wview daemon to be ready
    if (waitForWviewDaemon () == ERROR)
    {
        radMsgLog (PRI_HIGH, "waitForWviewDaemon failed");
        radMsgRouterExit ();
        procmonSysExit (&procmonWork);
        radProcessExit ();
        radSystemExit (WVIEW_SYSTEM_ID);
        exit (1);
    }

    //  get our configuration values
    if (stat (procmonWork.configFile, &fileStatus) == -1)
    {
        radMsgLog (PRI_CATASTROPHIC, "config file %s missing", 
                   procmonWork.configFile);
        radMsgRouterExit ();
        procmonSysExit (&procmonWork);
        radProcessExit ();
        radSystemExit (WVIEW_SYSTEM_ID);
        exit (1);
    }

    configFileId = radCfOpen (procmonWork.configFile);
    if (configFileId == NULL)
    {
        radMsgLog (PRI_CATASTROPHIC, "radCfOpen failed!\n");
        radMsgRouterExit ();
        procmonSysExit (&procmonWork);
        radProcessExit ();
        radSystemExit (WVIEW_SYSTEM_ID);
        exit (1);
    }

    // Populate our process table:
    memset (procmonWork.process, 0, sizeof(PMON_PROCESS) * PMON_PROCESS_MAX);
    for (i = 0; i < PMON_PROCESS_MAX; i ++)
    {
        if (radCfGetFirstEntry (configFileId, configIDs[i], instance, value) == ERROR)
        {
            // we can't do without this!
            radMsgLog (PRI_CATASTROPHIC, "PMON: %s: failed!", configIDs[i]);
            radCfClose (configFileId);
            radMsgRouterExit ();
            procmonSysExit (&procmonWork);
            radProcessExit ();
            radSystemExit (WVIEW_SYSTEM_ID);
            exit (1);
        }
        else
        {
            procmonWork.process[i].timeout = atoi (value);
            radMsgLog (PRI_STATUS, "PMON: %s: %d", 
                       configIDs[i], procmonWork.process[i].timeout);
        }

        if (procmonWork.process[i].timeout <= 0)
        {
            procmonWork.process[i].ticks = -1;
            procmonWork.process[i].pid = -1;
        }
        else
        {
            // Populate the rest of the values:
            strncpy (procmonWork.process[i].binFile, argv[0], 127);
            subString = strstr (procmonWork.process[i].binFile, "wvpmond");
            if (subString == NULL)
            {
                radMsgLog (PRI_CATASTROPHIC, "PMON: %s is invalid for bin path!", 
                           procmonWork.process[i].binFile);
                radCfClose (configFileId);
                radMsgRouterExit ();
                procmonSysExit (&procmonWork);
                radProcessExit ();
                radSystemExit (WVIEW_SYSTEM_ID);
                exit (1);
            }
            sprintf (subString, "%s", configIDs[i]);

            sprintf (procmonWork.process[i].pidFile, "%s/%s%s",
                     WVIEW_RUN_DIR, configIDs[i], ".pid");
            if (stat (procmonWork.process[i].pidFile, &fileStatus) == -1)
            {
                // Process is not running, mark it invalid
                radMsgLog (PRI_STATUS, "PMON: pid file %s not present, disable monitoring...",
                           procmonWork.process[i].pidFile);
                procmonWork.process[i].timeout = 0;
                procmonWork.process[i].ticks = -1;
                procmonWork.process[i].pid = -1;
            }
            else
            {
                // Get his pid from the pidfile
                procmonWork.process[i].pid = pmonGetProcessPid (procmonWork.process[i].pidFile);
                if (procmonWork.process[i].pid == ERROR)
                {
                    radMsgLog (PRI_CATASTROPHIC, "PMON: failed to READ %s!",
                               procmonWork.process[i].pidFile);
                    radCfClose (configFileId);
                    radMsgRouterExit ();
                    procmonSysExit (&procmonWork);
                    radProcessExit ();
                    radSystemExit (WVIEW_SYSTEM_ID);
                    exit (1);
                }
            }

            // Setup the state machine
            procmonWork.process[i].stateMachine = radStatesInit (&procmonWork);
            radStatesAddHandler (procmonWork.process[i].stateMachine,
                                 PMON_STATE_IDLE,
                                 IdleStateHandler);
            radStatesAddHandler (procmonWork.process[i].stateMachine,
                                 PMON_STATE_WAIT_RESP,
                                 WaitRespStateHandler);
            radStatesAddHandler (procmonWork.process[i].stateMachine,
                                 PMON_STATE_WAIT_EXIT,
                                 WaitExitStateHandler);
            radStatesAddHandler (procmonWork.process[i].stateMachine,
                                 PMON_STATE_WAIT_START,
                                 WaitStartStateHandler);
            radStatesSetState (procmonWork.process[i].stateMachine, PMON_STATE_IDLE);
        }
    }

    radCfClose (configFileId);


    radMsgRouterMessageRegister (WVIEW_MSG_TYPE_POLL_RESPONSE);

    // Start the timers:
    radProcessTimerStart (procmonWork.pollTimer, PMON_POLL_INTERVAL);
    radProcessTimerStart (procmonWork.tickTimer, PMON_TICK_INTERVAL);


    // enter normal processing
    procmonWork.inMainLoop = TRUE;
    radMsgLog (PRI_STATUS, "running...");
    
    while (!procmonWork.exiting)
    {
        // wait on something interesting
        if (radProcessWait (0) == ERROR)
        {
            procmonWork.exiting = TRUE;
        }
    }


    radMsgLog (PRI_STATUS, "exiting normally...");

    radMsgRouterExit ();
    radProcessTimerDelete (procmonWork.pollTimer);
    radProcessTimerDelete (procmonWork.tickTimer);
    procmonSysExit (&procmonWork);
    radProcessExit ();
    radSystemExit (WVIEW_SYSTEM_ID);
    exit (0);
}

